package PScm::Env;

use strict;
use warnings;
use base qw(PScm);

sub new {
    my ($class, %bindings) = @_;

    $class = ref($class) || $class;
    bless { bindings => {%bindings}, parent => 0 }, $class;
}

sub top {
    my ($self) = @_;
    if ($self->{parent}) {
        return $self->{parent}->top;
    } else {
        return $self;
    }
}

sub Extend {
    my ($self, $symbols, $values) = @_;

    return $self->ExtendUnevaluated($symbols,
                                    $values->map_eval($self));
}

sub ExtendUnevaluated {
    my ($self, $symbols, $values) = @_;

    my %bindings;
    $self->_populate_bindings(\%bindings, $symbols, $values);
    my $newenv = $self->new(%bindings);
    $newenv->{parent} = $self;
    return $newenv;
}

sub _populate_bindings {
    my ($self, $bindings, $symbols, $values) = @_;
    if ($symbols->is_null) {
        die "too many arguments\n" unless $values->is_null;
    } elsif ($symbols->is_pair) {
        if ($values->is_pair) {
            my $symbol = $symbols->first;
            if ($symbol->is_symbol) {
                $bindings->{$symbol->value} = $values->first;
                $self->_populate_bindings($bindings,
                                          $symbols->rest,
                                          $values->rest);
            } else {
                die "bad formal arguments[1]:",
                    $symbol->as_string(), "\n";
            }
        } else {
            die "not enough arguments\n";
        }
    } elsif ($symbols->is_symbol) {
        $bindings->{$symbols->value} = $values;
    } else {
        die "bad formal arguments[2]:",
            $symbols->as_string(), "\n";
    }
}

sub ExtendRecursively {
    my ($self, $symbols, $values) = @_;

    my $newenv = $self->ExtendUnevaluated($symbols, $values);
    $newenv->_eval_values();
    return $newenv;
}

sub ExtendIteratively {
    my ($self, $symbols, $values) = @_;

    if ($symbols->is_null) {
        return $self;
    } else {
        return $self->Extend(new PScm::Expr::List($symbols->first),
                             new PScm::Expr::List($values->first))
                    ->ExtendIteratively($symbols->rest,
                                        $values->rest);
    }
}

sub _eval_values {
    my ($self) = @_;
    map {
        $self->{bindings}{$_} =
          $self->{bindings}{$_}->Eval($self)
      }
      keys %{ $self->{bindings} };
}

sub LookUp {
    my ($self, $symbol) = @_;

    if (defined(my $ref = $self->_lookup_ref($symbol))) {
        return $$ref;
    } else {
        die "no binding for @{[$symbol->value]}",
          " in @{[ref($self)]}\n";
    }
}

sub Assign {
    my ($self, $symbol, $value) = @_;

    if (defined(my $ref = $self->_lookup_ref($symbol))) {
        $$ref = $value;
    } else {
        die "no binding for @{[$symbol->value]}",
          " in @{[ref($self)]}\n";
    }
}

sub _lookup_ref {
    my ($self, $symbol) = @_;

    if (exists($self->{bindings}{ $symbol->value })) {
        return \$self->{bindings}{ $symbol->value };
    } elsif ($self->{parent}) {
        return $self->{parent}->_lookup_ref($symbol);
    } else {
        return undef;
    }
}

1;
